"""Lists of topic-related icons."""

import re

GOOGLE_PATTERN = 'https://ssl.gstatic.com/calendar/images/eventillustrations/v1/img_%s_2x.jpg'
GOOGLE_TABLE = (
    # Language code: en
    ('new year', 'genericnewyear'),  # flag: 3
    ("new year's", 'genericnewyear'),  # flag: 3
    ('new years', 'genericnewyear'),  # flag: 3
    ('babyshower', 'babyshower'),  # flag: 2
    ('maternity', 'babyshower'),  # flag: 2
    ('baby shower', 'babyshower'),  # flag: 2
    ('jog', 'running'),  # flag: 2
    ('jogging', 'running'),  # flag: 2
    ('running', 'running'),  # flag: 2
    ('jogs', 'running'),  # flag: 2
    ('track and field', 'running'),  # flag: 2
    ('track & field', 'running'),  # flag: 2
    ('sprinting', 'running'),  # flag: 2
    ('book club', 'bookclub'),  # flag: 2
    ('reading', 'bookclub'),  # flag: 2
    ('thanksgiving', 'thanksgiving'),  # flag: 2
    ('date night', 'datenight'),  # flag: 3
    ('datenight', 'datenight'),  # flag: 3
    ('candlelight dinner', 'datenight'),  # flag: 3
    ('candle light dinner', 'datenight'),  # flag: 3
    ('romantic dinner', 'datenight'),  # flag: 3
    ('handicraft', 'handcraft'),  # flag: 2
    ('crochet', 'handcraft'),  # flag: 2
    ('knitting', 'handcraft'),  # flag: 2
    ('patchwork', 'handcraft'),  # flag: 2
    ('embroidery', 'handcraft'),  # flag: 2
    ('quilting', 'handcraft'),  # flag: 2
    ('sewing', 'handcraft'),  # flag: 2
    ('felting', 'handcraft'),  # flag: 2
    ('millinery', 'handcraft'),  # flag: 2
    ('cocktail', 'drinks'),  # flag: 2
    ('drinks', 'drinks'),  # flag: 2
    ('cocktails', 'drinks'),  # flag: 2
    ('stag night', 'drinks'),  # flag: 2
    ('hen night', 'drinks'),  # flag: 2
    ('bachelorette party', 'drinks'),  # flag: 2
    ('stag party', 'drinks'),  # flag: 2
    ('stag do', 'drinks'),  # flag: 2
    ('hen do', 'drinks'),  # flag: 2
    ('stag and doe', 'drinks'),  # flag: 2
    ('ladies night', 'drinks'),  # flag: 2
    ('wine night', 'drinks'),  # flag: 2
    ('wine bar', 'drinks'),  # flag: 2
    ('fridge repair', 'repair'),  # flag: 2
    ('handyman', 'repair'),  # flag: 2
    ('electrician', 'repair'),  # flag: 2
    ('plumber', 'repair'),  # flag: 2
    ('diy', 'repair'),  # flag: 2
    ('christmas', 'xmas'),  # flag: 3
    ('xmas', 'xmas'),  # flag: 3
    ('x-mas', 'xmas'),  # flag: 3
    ('boxing day', 'xmas'),  # flag: 3
    ('cooking', 'cooking'),  # flag: 3
    ('cook dinner', 'cooking'),  # flag: 3
    ('cook lunch', 'cooking'),  # flag: 3
    ('cook meal', 'cooking'),  # flag: 3
    ('prepare meal', 'cooking'),  # flag: 3
    ('prepare dinner', 'cooking'),  # flag: 3
    ('prepare lunch', 'cooking'),  # flag: 3
    ('make dinner', 'cooking'),  # flag: 3
    ('make lunch', 'cooking'),  # flag: 3
    ('kayaking', 'kayaking'),  # flag: 2
    ('canoeing', 'kayaking'),  # flag: 2
    ('canoe', 'kayaking'),  # flag: 2
    ('backtoschool', 'backtoschool'),  # flag: 2
    ('back to school', 'backtoschool'),  # flag: 2
    ('first school day', 'backtoschool'),  # flag: 2
    ('first day of school', 'backtoschool'),  # flag: 2
    ('back2school', 'backtoschool'),  # flag: 2
    ('back 2 school', 'backtoschool'),  # flag: 2
    ('equestrian', 'equestrian'),  # flag: 2
    ('dressage', 'equestrian'),  # flag: 2
    ('eventing', 'equestrian'),  # flag: 2
    ('horse riding', 'equestrian'),  # flag: 2
    ('horseriding', 'equestrian'),  # flag: 2
    ('santa claus', 'santa'),  # flag: 2
    ('father christmas', 'santa'),  # flag: 4
    ('wedding', 'wedding'),  # flag: 2
    ('wedding eve', 'wedding'),  # flag: 2
    ('wedding-eve party', 'wedding'),  # flag: 2
    ('weddings', 'wedding'),  # flag: 2
    ('quinceanera', 'quinceanera'),  # flag: 2
    ('fencing', 'fencing'),  # flag: 2
    ('violin', 'violin2'),  # flag: 2
    ('violins', 'violin2'),  # flag: 2
    ('manicure', 'manicure'),  # flag: 2
    ('pedicure', 'manicure'),  # flag: 2
    ('manicures', 'manicure'),  # flag: 2
    ('pedicures', 'manicure'),  # flag: 2
    ('yoga', 'yoga'),  # flag: 2
    ('going for a walk', 'walk'),  # flag: 2
    ('walking', 'walk'),  # flag: 2
    ('french course', 'learnlanguage'),  # flag: 2
    ('german course', 'learnlanguage'),  # flag: 2
    ('english course', 'learnlanguage'),  # flag: 2
    ('italian course', 'learnlanguage'),  # flag: 2
    ('chinese course', 'learnlanguage'),  # flag: 2
    ('japanese course', 'learnlanguage'),  # flag: 2
    ('korean course', 'learnlanguage'),  # flag: 2
    ('polish course', 'learnlanguage'),  # flag: 2
    ('spanish course', 'learnlanguage'),  # flag: 2
    ('arabic course', 'learnlanguage'),  # flag: 2
    ('hebrew course', 'learnlanguage'),  # flag: 2
    ('portuguese course', 'learnlanguage'),  # flag: 2
    ('thai course', 'learnlanguage'),  # flag: 2
    ('russian course', 'learnlanguage'),  # flag: 2
    ('turkish course', 'learnlanguage'),  # flag: 2
    ('dutch course', 'learnlanguage'),  # flag: 2
    ('bulgarian course', 'learnlanguage'),  # flag: 2
    ('greek course', 'learnlanguage'),  # flag: 2
    ('hindi course', 'learnlanguage'),  # flag: 2
    ('indonesian course', 'learnlanguage'),  # flag: 2
    ('vietnamese course', 'learnlanguage'),  # flag: 2
    ('norwegian course', 'learnlanguage'),  # flag: 2
    ('swedish course', 'learnlanguage'),  # flag: 2
    ('slovenian course', 'learnlanguage'),  # flag: 2
    ('ukranian course', 'learnlanguage'),  # flag: 2
    ('slovak course', 'learnlanguage'),  # flag: 2
    ('lithuanian course', 'learnlanguage'),  # flag: 2
    ('latvian course', 'learnlanguage'),  # flag: 2
    ('hungarian course', 'learnlanguage'),  # flag: 2
    ('finnish course', 'learnlanguage'),  # flag: 2
    ('filipino course', 'learnlanguage'),  # flag: 2
    ('farsi course', 'learnlanguage'),  # flag: 2
    ('danish course', 'learnlanguage'),  # flag: 2
    ('czech course', 'learnlanguage'),  # flag: 2
    ('croatian course', 'learnlanguage'),  # flag: 2
    ('catalan course', 'learnlanguage'),  # flag: 2
    ('french class', 'learnlanguage'),  # flag: 2
    ('german class', 'learnlanguage'),  # flag: 2
    ('english class', 'learnlanguage'),  # flag: 2
    ('italian class', 'learnlanguage'),  # flag: 2
    ('chinese class', 'learnlanguage'),  # flag: 2
    ('japanese class', 'learnlanguage'),  # flag: 2
    ('korean class', 'learnlanguage'),  # flag: 2
    ('polish class', 'learnlanguage'),  # flag: 2
    ('spanish class', 'learnlanguage'),  # flag: 2
    ('arabic class', 'learnlanguage'),  # flag: 2
    ('hebrew class', 'learnlanguage'),  # flag: 2
    ('portuguese class', 'learnlanguage'),  # flag: 2
    ('thai class', 'learnlanguage'),  # flag: 2
    ('russian class', 'learnlanguage'),  # flag: 2
    ('turkish class', 'learnlanguage'),  # flag: 2
    ('dutch class', 'learnlanguage'),  # flag: 2
    ('bulgarian class', 'learnlanguage'),  # flag: 2
    ('greek class', 'learnlanguage'),  # flag: 2
    ('hindi class', 'learnlanguage'),  # flag: 2
    ('indonesian class', 'learnlanguage'),  # flag: 2
    ('vietnamese class', 'learnlanguage'),  # flag: 2
    ('norwegian class', 'learnlanguage'),  # flag: 2
    ('swedish class', 'learnlanguage'),  # flag: 2
    ('slovenian class', 'learnlanguage'),  # flag: 2
    ('ukranian class', 'learnlanguage'),  # flag: 2
    ('slovak class', 'learnlanguage'),  # flag: 2
    ('lithuanian class', 'learnlanguage'),  # flag: 2
    ('latvian class', 'learnlanguage'),  # flag: 2
    ('hungarian class', 'learnlanguage'),  # flag: 2
    ('finnish class', 'learnlanguage'),  # flag: 2
    ('filipino class', 'learnlanguage'),  # flag: 2
    ('farsi class', 'learnlanguage'),  # flag: 2
    ('danish class', 'learnlanguage'),  # flag: 2
    ('czech class', 'learnlanguage'),  # flag: 2
    ('croatian class', 'learnlanguage'),  # flag: 2
    ('catalan class', 'learnlanguage'),  # flag: 2
    ('practice french', 'learnlanguage'),  # flag: 2
    ('practice german', 'learnlanguage'),  # flag: 2
    ('practice english', 'learnlanguage'),  # flag: 2
    ('practice italian', 'learnlanguage'),  # flag: 2
    ('practice chinese', 'learnlanguage'),  # flag: 2
    ('practice japanese', 'learnlanguage'),  # flag: 2
    ('practice korean', 'learnlanguage'),  # flag: 2
    ('practice polish', 'learnlanguage'),  # flag: 2
    ('practice spanish', 'learnlanguage'),  # flag: 2
    ('practice arabic', 'learnlanguage'),  # flag: 2
    ('practice hebrew', 'learnlanguage'),  # flag: 2
    ('practice portuguese', 'learnlanguage'),  # flag: 2
    ('practice thai', 'learnlanguage'),  # flag: 2
    ('practice russian', 'learnlanguage'),  # flag: 2
    ('practice turkish', 'learnlanguage'),  # flag: 2
    ('practice dutch', 'learnlanguage'),  # flag: 2
    ('practice bulgarian', 'learnlanguage'),  # flag: 2
    ('practice greek', 'learnlanguage'),  # flag: 2
    ('practice hindi', 'learnlanguage'),  # flag: 2
    ('practice indonesian', 'learnlanguage'),  # flag: 2
    ('practice vietnamese', 'learnlanguage'),  # flag: 2
    ('practice norwegian', 'learnlanguage'),  # flag: 2
    ('practice swedish', 'learnlanguage'),  # flag: 2
    ('practice slovenian', 'learnlanguage'),  # flag: 2
    ('practice ukranian', 'learnlanguage'),  # flag: 2
    ('practice slovak', 'learnlanguage'),  # flag: 2
    ('practice lithuanian', 'learnlanguage'),  # flag: 2
    ('practice latvian', 'learnlanguage'),  # flag: 2
    ('practice hungarian', 'learnlanguage'),  # flag: 2
    ('practice finnish', 'learnlanguage'),  # flag: 2
    ('practice filipino', 'learnlanguage'),  # flag: 2
    ('practice farsi', 'learnlanguage'),  # flag: 2
    ('practice danish', 'learnlanguage'),  # flag: 2
    ('practice czech', 'learnlanguage'),  # flag: 2
    ('practice croatian', 'learnlanguage'),  # flag: 2
    ('practice catalan', 'learnlanguage'),  # flag: 2
    ('reach out to', 'reachout'),  # flag: 2
    ('write letter', 'reachout'),  # flag: 2
    ('send invitations', 'reachout'),  # flag: 2
    ('videogames', 'videogaming'),  # flag: 2
    ('videogaming', 'videogaming'),  # flag: 2
    ('video games', 'videogaming'),  # flag: 2
    ('video game', 'videogaming'),  # flag: 2
    ('video gaming', 'videogaming'),  # flag: 2
    ('walk the dog', 'walkingdog'),  # flag: 2
    ('dogsitting', 'walkingdog'),  # flag: 2
    ('dog sitting', 'walkingdog'),  # flag: 2
    ('walk dog', 'walkingdog'),  # flag: 2
    ('dog walking', 'walkingdog'),  # flag: 2
    ('dog walker', 'walkingdog'),  # flag: 2
    ('take dog out', 'walkingdog'),  # flag: 2
    ('take out dog', 'walkingdog'),  # flag: 2
    ('basketball', 'basketball'),  # flag: 2
    ('sleeping', 'sleep'),  # flag: 2
    ('resting', 'sleep'),  # flag: 2
    ('relaxing', 'sleep'),  # flag: 2
    ('napping', 'sleep'),  # flag: 2
    ('sleep', 'sleep'),  # flag: 2
    ('nap', 'sleep'),  # flag: 2
    ('climbing', 'climbing'),  # flag: 2
    ('bouldering', 'climbing'),  # flag: 2
    ('massage', 'massage'),  # flag: 2
    ('back rub', 'massage'),  # flag: 2
    ('backrub', 'massage'),  # flag: 2
    ('massages', 'massage'),  # flag: 2
    ('bbq', 'bbq'),  # flag: 2
    ('barbecue', 'bbq'),  # flag: 2
    ('barbeque', 'bbq'),  # flag: 2
    ('dinner', 'dinner'),  # flag: 2
    ('dinners', 'dinner'),  # flag: 2
    ('restaurant', 'dinner'),  # flag: 2
    ('restaurants', 'dinner'),  # flag: 2
    ('family meal', 'dinner'),  # flag: 2
    ('american football', 'americanfootball'),  # flag: 2
    ('superbowl', 'americanfootball'),  # flag: 2
    ('super bowl', 'americanfootball'),  # flag: 2
    ('football', 'americanfootball'),  # flag: 2
    ('archery', 'archery'),  # flag: 2
    ('opera', 'theatreopera'),  # flag: 2
    ('theatre', 'theatreopera'),  # flag: 2
    ('theater', 'theatreopera'),  # flag: 2
    ('soccer', 'soccer'),  # flag: 2
    ('tennis', 'tennis'),  # flag: 2
    ('sail', 'sailing'),  # flag: 2
    ('sailing', 'sailing'),  # flag: 2
    ('boat cruise', 'sailing'),  # flag: 2
    ('sailboat', 'sailing'),  # flag: 2
    ('gym', 'gym'),  # flag: 2
    ('workout', 'gym'),  # flag: 2
    ('workouts', 'gym'),  # flag: 2
    ('fitness program', 'gym'),  # flag: 2
    ('fitness class', 'gym'),  # flag: 2
    ('fitness test', 'gym'),  # flag: 2
    ('fitness evaluation', 'gym'),  # flag: 2
    ('fitness center', 'gym'),  # flag: 2
    ('fitness training', 'gym'),  # flag: 2
    ('weightlifting', 'gym'),  # flag: 2
    ('weight lifting', 'gym'),  # flag: 2
    ('crossfit', 'gym'),  # flag: 2
    ('skiing', 'skiing2'),  # flag: 2
    ('ski', 'skiing2'),  # flag: 2
    ('skis', 'skiing2'),  # flag: 2
    ('snowboarding', 'skiing2'),  # flag: 2
    ('snowshoeing', 'skiing2'),  # flag: 2
    ('snow shoe', 'skiing2'),  # flag: 2
    ('snow boarding', 'skiing2'),  # flag: 2
    ('volleyball', 'volleyball'),  # flag: 2
    ('hiking', 'hiking'),  # flag: 2
    ('hike', 'hiking'),  # flag: 2
    ('hikes', 'hiking'),  # flag: 2
    ('mardi gras', 'mardigras'),  # flag: 2
    ('mardigras', 'mardigras'),  # flag: 2
    ('shrove tuesday', 'mardigras'),  # flag: 2
    ('fat tuesday', 'mardigras'),  # flag: 2
    ('coffee', 'coffee'),  # flag: 2
    ('coffees', 'coffee'),  # flag: 2
    ('learn to code', 'code'),  # flag: 2
    ('coding time', 'code'),  # flag: 2
    ('hackathon', 'code'),  # flag: 2
    ('rails girls', 'code'),  # flag: 2
    ('railsgirls', 'code'),  # flag: 2
    ('hour of code', 'code'),  # flag: 2
    ('codecademy', 'code'),  # flag: 2
    ('computer science', 'code'),  # flag: 2
    ('programming in python', 'code'),  # flag: 2
    ('web programming', 'code'),  # flag: 2
    ('programming in java', 'code'),  # flag: 2
    ('web development', 'code'),  # flag: 2
    ('dentist', 'dentist'),  # flag: 2
    ('dentistry', 'dentist'),  # flag: 2
    ('dental', 'dentist'),  # flag: 2
    ('teeth cleaning', 'dentist'),  # flag: 3
    ('golf', 'golf'),  # flag: 2
    ('triathlon', 'triathlon'),  # flag: 2
    ('islamic new year', 'islamicnewyear'),  # flag: 4
    ('hijri new year', 'islamicnewyear'),  # flag: 4
    ('parsi new year', 'islamicnewyear'),  # flag: 4
    ('dyke march', 'pride'),  # flag: 2
    ('christopher street day', 'pride'),  # flag: 2
    ('gay parade', 'pride'),  # flag: 2
    ('gay pride', 'pride'),  # flag: 2
    ('gayglers', 'pride'),  # flag: 2
    ('gaygler', 'pride'),  # flag: 2
    ('lesbian march', 'pride'),  # flag: 2
    ('lesbian parade', 'pride'),  # flag: 2
    ('lesbian pride', 'pride'),  # flag: 2
    ('euro pride', 'pride'),  # flag: 2
    ('europride', 'pride'),  # flag: 2
    ('world pride', 'pride'),  # flag: 2
    ('worldpride', 'pride'),  # flag: 2
    ('gay & lesbian', 'pride'),  # flag: 2
    ('gay lesbian', 'pride'),  # flag: 2
    ('gay and lesbian', 'pride'),  # flag: 2
    ('breakfast', 'breakfast'),  # flag: 2
    ('breakfasts', 'breakfast'),  # flag: 2
    ('brunch', 'breakfast'),  # flag: 2
    ('brunches', 'breakfast'),  # flag: 2
    ('haircut', 'haircut'),  # flag: 2
    ('hairdresser', 'haircut'),  # flag: 2
    ('hair', 'haircut'),  # flag: 2
    ('lunch', 'lunch'),  # flag: 2
    ('lunches', 'lunch'),  # flag: 2
    ('luncheon', 'lunch'),  # flag: 2
    ('reading', 'read'),  # flag: 2
    ('newspaper', 'read'),  # flag: 2
    ('ebook', 'read'),  # flag: 2
    ('persian new year', 'nowruz'),  # flag: 4
    ('nowruz', 'nowruz'),  # flag: 4
    ('artistic gymnastics', 'artisticgymnastics'),  # flag: 2
    ('boardgames', 'gamenight'),  # flag: 2
    ('boardgame', 'gamenight'),  # flag: 2
    ('board games', 'gamenight'),  # flag: 2
    ('board game', 'gamenight'),  # flag: 2
    ('ping pong', 'pingpong'),  # flag: 2
    ('table tennis', 'pingpong'),  # flag: 4
    ('ping-pong', 'pingpong'),  # flag: 2
    ('pingpong', 'pingpong'),  # flag: 2
    ('handball', 'handball'),  # flag: 2
    ('wrestling', 'wrestling'),  # flag: 2
    ('chinese new year', 'chinesenewyear'),  # flag: 4
    ('chinese new years', 'chinesenewyear'),  # flag: 4
    ("chinese new year's", 'chinesenewyear'),  # flag: 4
    ('vietnamese new year', 'chinesenewyear'),  # flag: 4
    ('tết', 'chinesenewyear'),  # flag: 4
    ('chinese new year', 'chinesenewyear'),  # flag: 4
    ('chinese lunar new year', 'chinesenewyear'),  # flag: 4
    ('rowing', 'rowing'),  # flag: 2
    ('head of the charles', 'rowing'),  # flag: 2
    ('may bumps', 'rowing'),  # flag: 2
    ('head of the river race', 'rowing'),  # flag: 2
    ('shot put', 'athleticsthrowing'),  # flag: 2
    ('discus', 'athleticsthrowing'),  # flag: 2
    ('javelin', 'athleticsthrowing'),  # flag: 2
    ('hammer throw', 'athleticsthrowing'),  # flag: 2
    ('st patricks', 'saintpatricksday'),  # flag: 2
    ("st patrick's", 'saintpatricksday'),  # flag: 2
    ('bicycle', 'cycling'),  # flag: 2
    ('cycling', 'cycling'),  # flag: 2
    ('bike', 'cycling'),  # flag: 2
    ('bicycles', 'cycling'),  # flag: 2
    ('bikes', 'cycling'),  # flag: 2
    ('biking', 'cycling'),  # flag: 2
    ('mountain biking', 'cycling'),  # flag: 2
    ('mountain bike', 'cycling'),  # flag: 2
    ('karate', 'karate'),  # flag: 2
    ('judo', 'karate'),  # flag: 2
    ('taekwondo', 'karate'),  # flag: 2
    ('martial arts', 'karate'),  # flag: 2
    ('jujutsu', 'karate'),  # flag: 2
    ('aikido', 'karate'),  # flag: 2
    ('jiu-jitsu', 'karate'),  # flag: 2
    ('jiu jitsu', 'karate'),  # flag: 2
    ('halloween', 'halloween'),  # flag: 2
    ('helloween', 'halloween'),  # flag: 2
    ("hallowe'en", 'halloween'),  # flag: 2
    ('allhalloween', 'halloween'),  # flag: 2
    ("all hallows' eve", 'halloween'),  # flag: 2
    ("all saints' eve", 'halloween'),  # flag: 2
    ('concert', 'concert'),  # flag: 2
    ('gig', 'concert'),  # flag: 2
    ('concerts', 'concert'),  # flag: 2
    ('gigs', 'concert'),  # flag: 2
    ('movie', 'cinema'),  # flag: 2
    ('cinema', 'cinema'),  # flag: 2
    ('movies', 'cinema'),  # flag: 2
    ('baseball', 'baseball'),  # flag: 2
    ('piano', 'learninstrument'),  # flag: 2
    ('singing', 'learninstrument'),  # flag: 2
    ('music class', 'learninstrument'),  # flag: 2
    ('choir practice', 'learninstrument'),  # flag: 2
    ('flute', 'learninstrument'),  # flag: 2
    ('orchestra', 'learninstrument'),  # flag: 2
    ('oboe', 'learninstrument'),  # flag: 2
    ('clarinet', 'learninstrument'),  # flag: 2
    ('saxophone', 'learninstrument'),  # flag: 2
    ('cornett', 'learninstrument'),  # flag: 2
    ('trumpet', 'learninstrument'),  # flag: 2
    ('contrabass', 'learninstrument'),  # flag: 2
    ('cello', 'learninstrument'),  # flag: 2
    ('trombone', 'learninstrument'),  # flag: 2
    ('tuba', 'learninstrument'),  # flag: 2
    ('music ensemble', 'learninstrument'),  # flag: 2
    ('string quartett', 'learninstrument'),  # flag: 2
    ('guitar lesson', 'learninstrument'),  # flag: 2
    ('classical music', 'learninstrument'),  # flag: 2
    ('choir', 'learninstrument'),  # flag: 2
    ('rhythmic gymnastics', 'rhythmicgymnastics'),  # flag: 2
    ('jumping', 'athleticsjumping'),  # flag: 2
    ('bowling', 'bowling'),  # flag: 2
    ('valentines day', 'valentinesday'),  # flag: 2
    ('valentine day', 'valentinesday'),  # flag: 2
    ("valentine's day", 'valentinesday'),  # flag: 2
    ('beer', 'beer'),  # flag: 2
    ('beers', 'beer'),  # flag: 2
    ('oktoberfest', 'beer'),  # flag: 2
    ('october fest', 'beer'),  # flag: 2
    ('octoberfest', 'beer'),  # flag: 2
    ('christmas party', 'xmasparty'),  # flag: 4
    ('xmas party', 'xmasparty'),  # flag: 4
    ('x-mas party', 'xmasparty'),  # flag: 4
    ('christmas eve party', 'xmasparty'),  # flag: 4
    ('xmas eve party', 'xmasparty'),  # flag: 4
    ('x-mas eve party', 'xmasparty'),  # flag: 4
    ('badminton', 'badminton'),  # flag: 2
    ('plan week', 'planmyday'),  # flag: 2
    ('plan quarter', 'planmyday'),  # flag: 2
    ('plan day', 'planmyday'),  # flag: 2
    ('plan vacation', 'planmyday'),  # flag: 2
    ('week planning', 'planmyday'),  # flag: 2
    ('vacation planning', 'planmyday'),  # flag: 2
    ('bmx', 'cyclingbmx'),  # flag: 2
    ('camping', 'camping'),  # flag: 2
    ('painting', 'art'),  # flag: 2
    ('art workshop', 'art'),  # flag: 2
    ('art workshops', 'art'),  # flag: 2
    ('sketching workshop', 'art'),  # flag: 2
    ('drawing workshop', 'art'),  # flag: 2
    ('dance', 'dancing'),  # flag: 2
    ('dancing', 'dancing'),  # flag: 2
    ('dances', 'dancing'),  # flag: 2
    ('ballet', 'dancing'),  # flag: 2
    ('water polo', 'waterpolo'),  # flag: 2
    ('waterpolo', 'waterpolo'),  # flag: 2
    ('christmas dinner', 'xmasmeal'),  # flag: 5
    ('christmas lunch', 'xmasmeal'),  # flag: 5
    ('christmas brunch', 'xmasmeal'),  # flag: 5
    ('christmas luncheon', 'xmasmeal'),  # flag: 5
    ('xmas dinner', 'xmasmeal'),  # flag: 5
    ('xmas lunch', 'xmasmeal'),  # flag: 5
    ('xmas brunch', 'xmasmeal'),  # flag: 5
    ('xmas luncheon', 'xmasmeal'),  # flag: 5
    ('x-mas dinner', 'xmasmeal'),  # flag: 5
    ('x-mas lunch', 'xmasmeal'),  # flag: 5
    ('x-mas brunch', 'xmasmeal'),  # flag: 5
    ('x-mas luncheon', 'xmasmeal'),  # flag: 5
    ('christmas eve dinner', 'xmasmeal'),  # flag: 5
    ('christmas eve lunch', 'xmasmeal'),  # flag: 5
    ('christmas eve brunch', 'xmasmeal'),  # flag: 5
    ('christmas eve luncheon', 'xmasmeal'),  # flag: 5
    ('xmas eve dinner', 'xmasmeal'),  # flag: 5
    ('xmas eve lunch', 'xmasmeal'),  # flag: 5
    ('xmas eve brunch', 'xmasmeal'),  # flag: 5
    ('xmas eve luncheon', 'xmasmeal'),  # flag: 5
    ('x-mas eve dinner', 'xmasmeal'),  # flag: 5
    ('x-mas eve lunch', 'xmasmeal'),  # flag: 5
    ('x-mas eve brunch', 'xmasmeal'),  # flag: 5
    ('x-mas eve luncheon', 'xmasmeal'),  # flag: 5
    ('clean the house', 'clean'),  # flag: 2
    ('clean the apartment', 'clean'),  # flag: 2
    ('clean house', 'clean'),  # flag: 2
    ('tidy up', 'clean'),  # flag: 2
    ('vacuum clean', 'clean'),  # flag: 2
    ('vacuum cleaning', 'clean'),  # flag: 2
    ('shooting sport', 'shooting'),  # flag: 2
    ('shooting sports', 'shooting'),  # flag: 2
    ('shooting competition', 'shooting'),  # flag: 2
    ('competitive shooting', 'shooting'),  # flag: 2
    ('graduation', 'graduation'),  # flag: 2
    ('oil change', 'oilchange'),  # flag: 2
    ('car service', 'oilchange'),  # flag: 2
    ('rugby', 'rugbysevens'),  # flag: 2
    ('swim', 'swimming'),  # flag: 2
    ('swimming', 'swimming'),  # flag: 2
    ('swims', 'swimming'),  # flag: 2
    ('diving', 'swimming'),  # flag: 2
    ('synchronized swimming', 'swimming'),  # flag: 2
    ('boxing', 'boxing'),  # flag: 2
    ('cricket game', 'cricket'),  # flag: 2
    ('cricket match', 'cricket'),  # flag: 2
    ('cricket competition', 'cricket'),  # flag: 2
    ('car repair', 'carmaintenance'),  # flag: 2
    ('auto repair', 'carmaintenance'),  # flag: 2
    ('car maintenance', 'carmaintenance'),  # flag: 2
    ('auto maintenance', 'carmaintenance'),  # flag: 2
    ('car mechanic', 'carmaintenance'),  # flag: 2
    ('auto mechanic', 'carmaintenance'),  # flag: 2
    ('tire replacement', 'carmaintenance'),  # flag: 2
    ('tire change', 'carmaintenance'),  # flag: 2
    ('field hockey', 'fieldhockey'),  # flag: 2
    ('billiard', 'billiard'),  # flag: 2
)


def match(name):
    """Given an event name like "Board Game Night!", return a URL to a relevant icon."""

    name = ' %s ' % name.lower()
    for table, base in ((GOOGLE_TABLE, GOOGLE_PATTERN),):
        for phrase, identifier in table:
            if re.search(re.escape(' %s ' % phrase).replace(r'\ ', r'\W+'), name):
                return base % identifier


if __name__ == '__main__':
    import os
    import sys

    from metabot.util import protobuf

    class _FlairPhrase(protobuf.ProtoBuf):  # pylint: disable=too-few-public-methods
        _fields = (
            (1, 'flag', int),
            (2, 'phrase', str),
        )

    class _FlairEntry(protobuf.ProtoBuf):  # pylint: disable=too-few-public-methods
        _fields = (
            (1, 'identifier', str),
            (2, 'phrases', _FlairPhrase, list),
        )

    class _FlairFile(protobuf.ProtoBuf):  # pylint: disable=too-few-public-methods
        _fields = (
            (1, 'entries', _FlairEntry, list),
            (2, 'lang_code', str),
        )

    def main(argv):  # pylint: disable=missing-docstring,too-many-branches,too-many-locals
        if len(argv) < 2:
            print('USAGE: %s path/to/assets/flairs/flairdata_en.pb' % sys.argv[0])
            return 1

        flair = _FlairFile(list(open(argv[1], 'rb').read()))

        lines = open(__file__, 'r').read().splitlines()
        with open(__file__ + '.new', 'w') as fobj:
            i = 0
            for i, line in enumerate(lines):
                fobj.write(line + '\n')
                if line == 'GOOGLE_TABLE = (':
                    break
            fobj.write('    # Language code: %s\n' % flair.lang_code)  # pylint: disable=no-member
            for identifier, phrases in flair.entries:  # pylint: disable=no-member
                for flag, phrase in phrases:
                    fobj.write('    %r,  # flag: %r\n' % ((phrase.lower(), identifier), flag))
            while lines[i] != ')':
                i += 1
            while i < len(lines):
                fobj.write(lines[i] + '\n')
                i += 1
        os.rename(__file__ + '.new', __file__)

    sys.exit(main(sys.argv))



"""Tests for metabot.modules.help."""

import pytest

from metabot.modules import help  # pylint: disable=redefined-builtin


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(help)


# pylint: disable=line-too-long


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Verify the help module."""

    assert conversation.message('/nothelp') == ''

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
I don't have much documentation—check with a bot admin!
"""

    assert conversation.message('/help') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/admin – Manage the admin list
"""

    assert conversation.message('/admin modulestestbot help hide admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › help: <b>Select a module</b>

<code>admin</code> is now hidden.
[Show admin (/admin – Manage the admin list) | /admin modulestestbot help unhide admin]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/help') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
I don't have much documentation—check with a bot admin!
"""

    assert conversation.message('/admin modulestestbot help hide admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › help: <b>Select a module</b>

<code>admin</code> is already hidden!
[Show admin (/admin – Manage the admin list) | /admin modulestestbot help unhide admin]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot help unhide admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › help: <b>Select a module</b>

<code>admin</code> is now visible.
[Hide admin (/admin – Manage the admin list) | /admin modulestestbot help hide admin]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot help unhide admin') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › help: <b>Select a module</b>

<code>admin</code> is not hidden!
[Hide admin (/admin – Manage the admin list) | /admin modulestestbot help hide admin]
[Back | /admin modulestestbot]
"""



"""Self-managing bot/module configuration store."""

import contextlib
import logging
import os

from metabot.util import dicttools
from metabot.util import jsonutil
from metabot.util import yamlutil


class BotConf(dicttools.ImplicitTrackingDict):
    """Self-managing bot/module configuration store."""

    confdir = None

    def __init__(self, confdir=None):  # pylint: disable=too-many-branches
        super().__init__()
        if confdir:  # pylint: disable=too-many-nested-blocks
            self.confdir = confdir
            if not os.path.isdir(confdir):
                os.makedirs(confdir, 0o700)  # pragma: no cover
            else:
                for fname in os.listdir(confdir):
                    if fname.endswith('.yaml'):
                        self[fname[:-len('.yaml')]] = yamlutil.load(os.path.join(confdir, fname))

            # Schema update: Remove after 2019-03-26 (https://github.com/nmlorg/metabot/issues/18).
            if not self['bots']:
                fname = os.path.join(confdir, 'multibot.json')
                data = jsonutil.load(fname)
                if data:  # pragma: no cover
                    self['bots'] = data
                    logging.info('Converted %s to %s.', fname, os.path.join(confdir, 'bots.yaml'))

            # Schema update: Remove after 2019-04-05 (https://github.com/nmlorg/metabot/issues/18).
            for botconf in self['bots'].values():
                for groupid, groupconf in botconf['moderator'].items():  # pragma: no cover
                    groupid = int(groupid)
                    for k in ('title', 'type', 'username'):
                        if k in groupconf:
                            self['groups'][groupid][k] = groupconf[k]

            # Schema update: Remove after 2019-06-12 (https://github.com/nmlorg/metabot/issues/37).
            for botconf in self['bots'].values():
                if botconf.get('telegram'):  # pragma: no cover
                    for modname in list(botconf):
                        if modname != 'issue37':
                            botconf['issue37'][modname] = botconf[modname]
                            botconf.pop(modname)

            # Schema update: Remove after 2019-08-13 (https://github.com/nmlorg/metabot/issues/41).
            for botconf in self['bots'].values():
                for command, message in botconf['issue37']['echo'].items():
                    if not isinstance(message, dict):  # pragma: no cover
                        botconf['issue37']['echo'][command] = {
                            'text': message,
                            'paginate': True,
                            'private': '\n' in message,
                        }

            # Schema update: Remove after 2020-06-22 (https://github.com/nmlorg/metabot/issues/8).
            for botconf in self['bots'].values():  # pragma: no cover
                for groupconf in list(botconf['issue37']['moderator'].values()):
                    dailyconf = groupconf.get('daily')
                    if dailyconf is not None and not isinstance(dailyconf, dict):
                        groupconf['daily'] = {'hour': dailyconf}
                        for key in ('dow', 'text'):
                            value = groupconf.pop('daily' + key)
                            if value is not None:
                                groupconf['daily'][key] = value

        self._fnames = set(self)

    @contextlib.contextmanager
    def record_mutations(self, ctx):
        """Capture and record all changes to the bot config."""

        try:
            yield self
        finally:
            log = self.finalize()
            if log:
                userdata = []
                if ctx.userinfo:
                    userdata.append('%s' % ctx.user['id'])
                    if ctx.userinfo.data.username:
                        userdata.append('@' + ctx.userinfo.data.username)
                    if ctx.userinfo.data.name:
                        userdata.append(repr(ctx.userinfo.data.name))
                for path, (value, orig) in sorted(log.items()):
                    pathstr = '.'.join('%s' % part for part in path)
                    logging.info('[%s] %s: %r -> %r', ' '.join(userdata), pathstr, orig, value)
                self.save()

    def save(self):
        """Serialize the store to disk (if confdir was provided at creation)."""

        if self.confdir:
            for fname in self._fnames.difference(self):
                os.remove(os.path.join(self.confdir, fname + '.yaml'))  # pragma: no cover
            self._fnames = set(self)
            for fname, data in self.items():
                yamlutil.dump(os.path.join(self.confdir, fname + '.yaml'), data)



"""Modularized, multi-account bot."""

import logging

from metabot.modules import admin
from metabot import multibot
from metabot.util import humanize
from metabot.util import modutil


def main():  # pylint: disable=missing-docstring
    logging.basicConfig(format='%(asctime)s %(levelname)s %(filename)s:%(lineno)s] %(message)s',
                        level=logging.INFO)

    mybot = multibot.MultiBot(modutil.load_modules('metabot.modules'), confdir='config')
    if not mybot.conf['bots']:
        print()
        print("Hi! Before I can start, I need at least one bot's Telegram token. If you don't have "
              'one already, follow the instructions at:')
        print()
        print('    https://core.telegram.org/bots#creating-a-new-bot')
        print()
        print('At the end, BotFather will send you a message starting with "Use this token" '
              'followed by a line starting with a number. Copy the full line starting with the '
              'number and paste it here:')
        print()
        while not mybot.conf['bots']:
            initial_token = input('Telegram token: ')
            try:
                username = mybot.add_bot(initial_token)
            except AssertionError:
                print()
                print("That doesn't look quite right. Look for a message from BotFather starting "
                      'with "Use this token" followed by a line starting with a number. Copy the '
                      'full line starting with the number and paste it here:')
                print()
            except Exception as exc:  # pylint: disable=broad-except
                print()
                print('Woops, that generated: %r', exc)
            else:
                mybot.run_bot(username)
    unconfigured = sorted(username for username, botconf in mybot.conf['bots'].items()
                          if not botconf['issue37']['admin'].get('admins'))
    if unconfigured:
        print()
        print('To configure %s, open a chat with:' % humanize.list(unconfigured))
        print()
        for username in unconfigured:
            print('    https://t.me/%s' % username)
        print()
        print('and type:')
        print()
        print('    /_bootstrap %s' % admin.BOOTSTRAP_TOKEN)
        print()
    mybot.run()


if __name__ == '__main__':
    main()



"""Count down to a specific date/time."""

import time

from metabot.util import adminui
from metabot.util import humanize


def modhelp(unused_ctx, modconf, sections):  # pylint: disable=missing-docstring
    now = time.time()
    for command, timestamp in modconf.items():
        if now > timestamp:
            sections['commands'].add('/%s \u2013 Count up from %s' % (command, timestamp))
        else:
            sections['commands'].add('/%s \u2013 Count down to %s' % (command, timestamp))


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query') and ctx.command in modconf:
        return countdown(msg, modconf[ctx.command])

    return False


def countdown(msg, timestamp):  # pylint: disable=missing-docstring
    msg.add(format_delta(timestamp - time.time()))


def format_delta(seconds):
    """Format a number of seconds into "5 days, 1 hour, 13.4 seconds", etc."""

    if seconds < 0:
        suffix = ' ago'
        seconds = -seconds
    else:
        suffix = ''
    days, seconds = divmod(seconds, 60 * 60 * 24)
    hours, seconds = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(seconds, 60)
    seconds = round(seconds, 2)
    if seconds == round(seconds):
        seconds = int(seconds)
    message = []
    if days:
        message.append(humanize.plural(days, 'day', '<b>%i</b> %s'))
    if hours:
        message.append(humanize.plural(hours, 'hour', '<b>%i</b> %s'))
    if minutes:
        message.append(humanize.plural(minutes, 'minute', '<b>%i</b> %s'))
    if seconds:
        message.append(humanize.plural(seconds, 'second', '<b>%s</b> %s'))
    if message:
        return ', '.join(message) + suffix
    return '<b>NOW</b>'


def admin(frame):
    """Handle /admin BOTNAME countdown."""

    msg = frame.msg
    menu = adminui.Menu()
    for command in frame.value:
        menu.add(command)
    menu.add(None)
    newframe, handler = menu.select(frame)
    if handler:
        if newframe.text.isdigit():
            adminui.set_log(newframe, int(newframe.text))
        elif newframe.text.lower() in ('-', 'none', 'off', 'remove'):
            adminui.set_log(newframe, None)
        else:
            if newframe.text:
                msg.add("I'm not sure how to count down to <code>%s</code>!", newframe.text)

            msg.path(newframe.field)
            msg.action = 'Type the time for /' + newframe.field
            msg.add('This is a little technical (it will be made simpler in the future), but type '
                    'the unix timestamp to count down to.')
            msg.add('(Go to https://www.epochconverter.com/, fill out the section "Human date to '
                    'Timestamp", then use the number listed next to "Epoch timestamp".)')
            if newframe.value:
                msg.add('To remove /%s (which is counting to %s), type "off".', newframe.field,
                        newframe.value)
            return

    msg.action = 'Choose a command'
    msg.add("Type the name of a command to add (like <code>days</code>\u2014don't include a slash "
            'at the beginning!), or select an existing countdown to remove.')
    # See https://github.com/nmlorg/metabot/issues/65.
    menu = adminui.Menu()
    for command in frame.value:
        menu.add(command)
    menu.display(frame, what='command')



"""Tests for metabot.modules.countdown."""

import pytest

from metabot.modules import countdown


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(countdown)


# pylint: disable=line-too-long


def test_countdown(conversation):  # pylint: disable=redefined-outer-name
    """Verify the countdown module (which uses dynamic commands)."""

    assert conversation.message('/mycountdown') == ''

    conversation.bot.config['issue37']['countdown']['mycountdown'] = 1534906800
    assert conversation.message('/mycountdown').endswith(' ago\n')

    conversation.bot.config['issue37']['countdown']['mycountdown'] = 15349068000
    assert not conversation.message('/mycountdown').endswith(' ago\n')


def test_format_delta():
    """Verify the time delta formatter."""

    assert countdown.format_delta(0.) == '<b>NOW</b>'
    assert countdown.format_delta(1 / 9) == '<b>0.11</b> seconds'
    assert countdown.format_delta(1.5) == '<b>1.5</b> seconds'
    assert countdown.format_delta(61.) == '<b>1</b> minute, <b>1</b> second'
    assert countdown.format_delta(
        ((365 * 24 + 1) * 60 + 2) * 60.) == '<b>365</b> days, <b>1</b> hour, <b>2</b> minutes'


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Test /help."""

    conversation.bot.config['issue37']['countdown']['count1'] = 1534906800
    conversation.bot.config['issue37']['countdown']['count2'] = 15349068000

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/count1 – Count up from 1534906800

/count2 – Count down to 15349068000
"""


def test_admin(conversation):  # pylint: disable=redefined-outer-name
    """Test /admin BOTNAME countdown."""

    assert conversation.message('/admin modulestestbot countdown') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown: <b>Choose a command</b>

Type the name of a command to add (like <code>days</code>—don't include a slash at the beginning!), or select an existing countdown to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('CountDownTest') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown › countdowntest: <b>Type the time for /countdowntest</b>

This is a little technical (it will be made simpler in the future), but type the unix timestamp to count down to.

(Go to https://www.epochconverter.com/, fill out the section "Human date to Timestamp", then use the number listed next to "Epoch timestamp".)
[Back | /admin modulestestbot countdown]
"""

    assert conversation.message('1534906800') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown: <b>Choose a command</b>

Set <code>countdowntest</code> to <code>1534906800</code>.

Type the name of a command to add (like <code>days</code>—don't include a slash at the beginning!), or select an existing countdown to remove.
[countdowntest (1534906800) | /admin modulestestbot countdown countdowntest]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot countdown countdowntest 1000') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown: <b>Choose a command</b>

Changed <code>countdowntest</code> from <code>1534906800</code> to <code>1000</code>.

Type the name of a command to add (like <code>days</code>—don't include a slash at the beginning!), or select an existing countdown to remove.
[countdowntest (1000) | /admin modulestestbot countdown countdowntest]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot countdown countdowntest bogus<>') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown › countdowntest: <b>Type the time for /countdowntest</b>

I'm not sure how to count down to <code>bogus&lt;&gt;</code>!

This is a little technical (it will be made simpler in the future), but type the unix timestamp to count down to.

(Go to https://www.epochconverter.com/, fill out the section "Human date to Timestamp", then use the number listed next to "Epoch timestamp".)

To remove /countdowntest (which is counting to 1000), type "off".
[Back | /admin modulestestbot countdown]
"""

    assert conversation.message('remove') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown: <b>Choose a command</b>

Unset <code>countdowntest</code> (was <code>1000</code>).

Type the name of a command to add (like <code>days</code>—don't include a slash at the beginning!), or select an existing countdown to remove.
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot countdown bogus remove') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › countdown: <b>Choose a command</b>

<code>bogus</code> is already unset.

Type the name of a command to add (like <code>days</code>—don't include a slash at the beginning!), or select an existing countdown to remove.
[Back | /admin modulestestbot]
"""



"""Tests for metabot.calendars.multicalendar."""

import collections

from metabot.calendars import multicalendar


def test_multicalendar(monkeypatch):
    """Run through most of the functionality of multicalendar.MultiCalendar."""

    multical = multicalendar.MultiCalendar()
    assert multical.calendars == {}
    assert multical.by_local_id == {}
    assert multical.ordered == []

    events = {
        'alpha': {
            'end': 3000,
            'local_id': 'alpha',
            'start': 2000,
            'summary': 'Alpha',
        },
        'bravo': {
            'end': 7000,
            'local_id': 'bravo',
            'start': 5000,
            'summary': 'Bravo',
        },
        'charlie': {
            'end': 8000,
            'local_id': 'charlie',
            'start': 6000,
            'summary': 'Charlie',
        },
    }
    dummycal = collections.namedtuple('_', 'events poll')(events, lambda: True)

    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        multical.add('dummy:dummy')

    assert multical.by_local_id == {'alpha': 0, 'bravo': 1, 'charlie': 2}
    assert multical.get_event('unset') == (None, None, None)
    assert multical.get_event('alpha') == (None, events['alpha'], events['bravo'])
    assert multical.get_event('bravo') == (events['alpha'], events['bravo'], events['charlie'])
    assert multical.get_event('charlie') == (events['bravo'], events['charlie'], None)

    #            |2000 3000|
    # |1000 1999|
    assert list(multical.get_overlap(1000, 1999)) == []

    #            |2000 3000|
    # |1000  2000|
    assert list(multical.get_overlap(1000, 2000)) == [events['alpha']]

    #            |2000 3000|
    # |1000                       4000|
    assert list(multical.get_overlap(1000, 4000)) == [events['alpha']]

    #            |2000 3000|
    #                      |3000  4000|
    assert list(multical.get_overlap(3000, 4000)) == [events['alpha']]

    #            |2000 3000|
    #                       |3001 4000|
    assert list(multical.get_overlap(3001, 4000)) == []

    assert list(multical.get_overlap(5000, 8000)) == [events['bravo'], events['charlie']]

    by_local_id = multical.by_local_id
    ordered = multical.ordered
    assert multical.poll()
    assert multical.by_local_id is not by_local_id and multical.by_local_id == by_local_id
    assert multical.ordered is not ordered and multical.ordered == ordered


def test_current_index(monkeypatch):  # pylint: disable=too-many-statements
    """Test the specific behavior of _current_index and current_index."""

    events = {
        'alpha': {
            'end': 3000,
            'local_id': 'alpha',
            'start': 2000,
            'summary': 'Alpha',
        },
        'bravo': {
            'end': 7000,
            'local_id': 'bravo',
            'start': 5000,
            'summary': 'Bravo',
        },
        'charlie': {
            'end': 8000,
            'local_id': 'charlie',
            'start': 6000,
            'summary': 'Charlie',
        },
    }
    dummycal = collections.namedtuple('_', 'events')(events)

    # pylint: disable=protected-access

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 1000.)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['alpha']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 3000.)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['alpha']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 3000.5)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['bravo']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 7000.)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['bravo']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 7000.5)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['charlie']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 8000.)
        multical.add('dummy:dummy')
    assert multical._current_index == multical.by_local_id['charlie']

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 8000.5)
        multical.add('dummy:dummy')
    assert multical._current_index is None

    multical = multicalendar.MultiCalendar()
    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal)
        monkey.setattr('time.time', lambda: 1000.)
        multical.add('dummy:dummy')
        assert multical.current_index == multical.by_local_id['alpha']
        assert multical.current_local_id == 'alpha'
        monkey.setattr('time.time', lambda: 3000.)
        assert multical.current_local_id == 'alpha'
        monkey.setattr('time.time', lambda: 3000.5)
        assert multical.current_local_id == 'bravo'
        monkey.setattr('time.time', lambda: 8000.)
        assert multical.current_local_id == 'charlie'
        assert multical.get_event() == (events['bravo'], events['charlie'], None)
        monkey.setattr('time.time', lambda: 8000.5)
        assert multical.current_index is None
        assert multical.current_local_id is None
        assert multical.get_event() == (None, None, None)


def test_view(monkeypatch):
    """Run through the behavior of multicalendar.View (via multicalendar.MultiCalendar.view)."""

    multical = multicalendar.MultiCalendar()

    calendars = {
        'xray': {
            'xray:alpha': {
                'end': 3000,
                'local_id': 'xray:alpha',
                'start': 2000,
                'summary': 'Alpha',
            },
        },
        'yankee': {
            'yankee:bravo': {
                'end': 7000,
                'local_id': 'yankee:bravo',
                'start': 5000,
                'summary': 'Bravo',
            },
        },
        'zulu': {
            'zulu:charlie': {
                'end': 8000,
                'local_id': 'zulu:charlie',
                'start': 6000,
                'summary': 'Charlie',
            },
        },
    }
    dummycal = collections.namedtuple('_', 'events')

    with monkeypatch.context() as monkey:
        monkey.setattr('metabot.calendars.loader.get', lambda calid: dummycal(calendars[calid]))
        monkey.setattr('time.time', lambda: 1000.)
        multical.add('xray')
        multical.add('yankee')
        multical.add('zulu')

    assert multical.by_local_id == {'xray:alpha': 0, 'yankee:bravo': 1, 'zulu:charlie': 2}
    alpha, bravo, charlie = multical.ordered  # pylint: disable=unbalanced-tuple-unpacking

    assert multical.get_event('xray:alpha') == (None, alpha, bravo)
    assert multical.get_event('yankee:bravo') == (alpha, bravo, charlie)
    assert multical.get_event('zulu:charlie') == (bravo, charlie, None)

    view = multical.view({'xray'})

    assert view.get_event('unset') == (None, None, None)
    assert view.get_event('xray:alpha') == (None, alpha, None)

    view = multical.view({'xray', 'zulu'})

    assert view.get_event('xray:alpha') == (None, alpha, charlie)
    assert view.get_event('zulu:charlie') == (alpha, charlie, None)

    assert list(multical.get_overlap(3000, 6000)) == [alpha, bravo, charlie]
    assert list(view.get_overlap(3000, 6000)) == [alpha, charlie]

    with monkeypatch.context() as monkey:
        monkey.setattr('time.time', lambda: 1000.)
        assert multical.current_index == 0
        assert view.current_index == 0
        assert view.get_event() == (None, alpha, charlie)

        monkey.setattr('time.time', lambda: 4000.)
        assert multical.current_index == 1
        assert view.current_index == 2
        assert view.get_event() == (alpha, charlie, None)

        view = multical.view({'xray'})

        assert multical.current_index == 1
        assert view.current_index is None
        assert view.get_event() == (None, None, None)

        monkey.setattr('time.time', lambda: 9000.)
        assert multical.current_index is None
        assert view.current_index is None
        assert view.get_event() == (None, None, None)



"""Quick routine to extract ticket information from event descriptions."""

import json
import logging
import re

import requests


def get_info(text):
    """Quick routine to extract ticket information from event descriptions."""

    for provider in (_brownpapertickets, _eventbrite):
        yield from provider(text)


def _brownpapertickets(text):
    events = {
        event_id for _, _, event_id in re.findall(
            '://(www.|m.|)(brownpapertickets.com|bpt.me)/event/([0-9]+)', text)
    }
    for event_id in sorted(events):
        ret = re.search('>([0-9]+) tickets remaining',
                        requests.get('https://m.bpt.me/event/' + event_id, timeout=10).text)
        if ret:
            remaining = int(ret.groups()[0])
        else:
            remaining = 0
        yield remaining, 'https://bpt.me/event/' + event_id


def _eventbrite(text):
    events = {
        event_id for _, event_id in re.findall('://www.eventbrite.com/e/([^/]+-)?([0-9]+)', text)
    }
    for event_id in sorted(events):
        url = 'https://www.eventbrite.com/ajax/event/%s/ticket_classes/for_sale/' % event_id
        try:
            data = requests.get(url,
                                timeout=10,
                                params={
                                    'pos': 'online',
                                },
                                headers={
                                    'x-requested-with': 'XMLHttpRequest',
                                }).json()
        except json.decoder.JSONDecodeError:
            logging.exception('Decoding %s:', url)
        else:
            remaining = 0
            for ticket_class in data['ticket_classes']:
                if ticket_class['number_of_tickets_remaining']:  # If unset, this is None, not 0.
                    remaining += ticket_class['number_of_tickets_remaining']
            yield remaining, data['event_info']['vanity_url']



"""Display recent and upcoming events."""

import datetime
import logging
import time
import urllib.parse

import pytz

from metabot.util import geoutil
from metabot.util import html
from metabot.util import humanize
from metabot.util import icons
from metabot.util import tickets

# https://www.weather.gov/forecast-icons
BAD_WEATHER_KEYWORDS = {
    'blizzard', 'freezing', 'heavy', 'hurricane', 'ice', 'snow', 'tornado', 'tropical'
}


def get_group_conf(groupconf):
    """Pull calendar configuration from a raw group conf."""

    timezone = groupconf.get('timezone')
    tzinfo = timezone and pytz.timezone(timezone)
    return (groupconf.get('calendars', '').split(), tzinfo, groupconf.get('maxeventscount', 10),
            groupconf.get('maxeventsdays',
                          6), groupconf['daily'].get('hour'), groupconf['daily'].get('dow', 0))


def get_group_events(bot, calcodes, tzinfo, count, days, now=None):  # pylint: disable=too-many-arguments,too-many-locals
    """Build lists of events and weather alerts for the given calendar configuration."""

    calendar_view = bot.multibot.multical.view(calcodes)
    if now is None:
        now = time.time()
    nowdt = datetime.datetime.fromtimestamp(now, tzinfo)
    midnight = nowdt.replace(hour=0, minute=0, second=0, microsecond=0)
    period = (midnight + datetime.timedelta(days=days + 1) - nowdt).total_seconds()
    all_events = list(calendar_view.get_overlap(now, now + max(period, 60 * 60 * 24 * 100)))
    alerts = {}
    for event in all_events:
        if event['location']:
            try:
                alertlist = geoutil.weatheralerts(event['location'])
            except Exception:  # pylint: disable=broad-except
                logging.exception('Ignoring:')
            else:
                for alert in alertlist or ():
                    alerts[alert['id']] = alert
    end = now + period
    return ([event for event in all_events[:count] if event['start'] <= end],
            [alert for alertid, alert in sorted(alerts.items())])


def format_event(bot, event, tzinfo, full=True):
    """Given a metabot.calendars.base.Calendar event, build a human-friendly representation."""

    message = '<b>%s</b>' % event['summary']
    for count, url in tickets.get_info(event['description']):
        if count:
            message = '%s [<a href="%s">%i tickets remaining</a>]' % (message, url, count)
        else:
            message = '%s [<a href="%s">tickets</a>]' % (message, url)
    message = '%s\n<a href="%s">%s</a>' % (message,
                                           bot.encode_url('/events %s %s' %
                                                          (event['local_id'], tzinfo.zone)),
                                           humanize_range(event['start'], event['end'], tzinfo))
    if event['location']:
        location_name = event['location'].split(',', 1)[0]
        location_url = 'https://maps.google.com/maps?' + urllib.parse.urlencode({
            'q': event['location'].encode('utf-8'),
        })  # yapf: disable
        message = '%s @ <a href="%s">%s</a>' % (message, location_url, location_name)
        geo = format_geo(event['location'], event['start'])
        if geo:
            message = '%s\n\u26a0 %s' % (message, geo)
    if full and event['description']:
        message = '%s\n\n%s' % (message, html.sanitize(event['description']))
    return message


def format_geo(address, now):
    """Build a string of weather exceptions for the given address as of the given time."""

    try:
        forecast = geoutil.hourlyforecast(address, now)
    except Exception:  # pylint: disable=broad-except
        logging.exception('Ignoring:')
        return
    if not forecast:
        return
    warnings = []
    if set(forecast['shortForecast'].lower().split()) & BAD_WEATHER_KEYWORDS:
        warnings.append(forecast['shortForecast'])
    if warnings:
        if forecast['temperatureUnit'] == 'F':
            warnings.append('%s\u2109' % forecast['temperature'])
        elif forecast['temperatureUnit'] == 'C':
            warnings.append('%s\u2103' % forecast['temperature'])
        return ' \u2022 '.join(warnings)


def humanize_range(start, end, tzinfo):
    """Return the range between start and end as human-friendly text."""

    return humanize.range(datetime.datetime.fromtimestamp(start, tzinfo),
                          datetime.datetime.fromtimestamp(end, tzinfo))


def get_image(event, botconf):
    """Choose the best image for a given event and botconf."""

    eventconf = botconf['issue37']['events']
    eventimage = eventconf['events'].get(event['local_id'])
    if eventimage:
        return eventimage
    for pattern, seriesimage in eventconf['series'].items():
        if pattern.lower() in event['summary'].lower():
            return seriesimage
    return icons.match(event['summary']) or icons.match(event['description'])



"""Process natural-language messages."""

import logging
import re

import pytz

from metabot.util import eventutil


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring,unused-argument
    if ctx.type == 'message' and not ctx.command:
        return process(ctx, msg)

    return False


PATTERNS = ('when is( the( next)?)? (?P<event>.+?)( (at|in)( the)? (?P<location>.+?))?',)


def process(ctx, msg):
    """Look for free-form text corresponding to a line in PATTERNS, and dispatch it."""

    text = ' '.join(re.split(r'\W+', ctx.text.lower())).strip()

    for pattern in PATTERNS:
        ret = re.match(pattern + '$', text)
        if not ret:
            continue
        fields = ret.groupdict()
        logging.info('Parsed: %s', ', '.join('%s=%r' % (k, v) for k, v in sorted(fields.items())))
        if fields.get('event'):
            return process_event(ctx, msg, fields)

    return False


def process_event(ctx, msg, query):  # pylint: disable=too-many-branches,too-many-locals
    """Process queries of the form "When is the next XXX?"."""

    if ctx.chat['type'] == 'private':
        conf = ctx.bot.config['issue37']['events']['users']['%s' % ctx.user['id']]
    else:
        conf = ctx.bot.config['issue37']['moderator']['%s' % ctx.chat['id']]
    calcodes = conf.get('calendars', '').split()
    timezone = pytz.timezone(conf.get('timezone', 'UTC'))

    did_first = did_cal = did_summary = did_venue = False
    summaries = set()
    venues = set()
    local_id = None
    while not did_first or not did_cal or not did_summary or not did_venue:
        _, event, nextev = ctx.bot.multibot.multical.get_event(local_id)
        if not event:
            break
        summary = event['summary'].lower()
        location = event['location'].lower()
        if query['event'] in summary and (query.get('location') or '') in location:
            venue = location.split(',', 1)[0]
            reasons = set()
            calcode = event['local_id'].split(':', 1)[0]
            if calcode not in calcodes:
                if not did_first:
                    did_first = True
                    reasons.add(query['event'] + ' I know about')
            elif not did_first:
                did_first = did_cal = True
                reasons.add(query['event'])
            elif not did_cal:
                did_cal = True
                reasons.add('%s on the %s calendar' %
                            (query['event'], ctx.bot.multibot.calendars[calcode]['name']))
            else:
                if not did_summary and summary not in summaries:
                    did_summary = True
                    reasons.add(event['summary'])
                if not did_venue and venue not in venues:
                    did_venue = True
                    reasons.add('%s at %s' % (query['event'], event['location'].split(',', 1)[0]))

            if reasons:
                msg.add('<i>The next %s is:</i>', ' / '.join(sorted(reasons)))
                msg.add(eventutil.format_event(ctx.bot, event, timezone, full=False))
                summaries.add(summary)
                venues.add(venue)

        if not nextev:
            break
        local_id = nextev['local_id']



"""Test environment defaults."""

import json

import ntelebot
from ntelebot.conftest import _bot_mock  # pylint: disable=unused-import
import pytest

from metabot.modules import admin
from metabot.modules import help  # pylint: disable=redefined-builtin
from metabot import multibot


@pytest.fixture(autouse=True)
def _dont_mangle_callback_data(monkeypatch):
    monkeypatch.setattr('ntelebot.keyboardutil.fix', lambda keyboard, maxlen=0: None)


@pytest.fixture(autouse=True)
def _disable_geoutil(monkeypatch):
    monkeypatch.setattr('metabot.util.geoutil._CLIENT', None)


def _format_message(response):
    text = response.pop('text', None) or response.pop('caption', '(EMPTY MESSAGE)')
    reply_markup = response.pop('reply_markup', None)
    header = ' '.join('%s=%s' % (field, value) for field, value in sorted(response.items()))
    text = '[%s]\n%s\n' % (header, text)
    if reply_markup:
        for row in reply_markup.pop('inline_keyboard', ()):
            text = '%s%s\n' % (text, ' '.join(
                '[%s | %s]' % (button['text'], button['callback_data']) for button in row))
        if reply_markup:  # pragma: no cover
            text = '%s%s\n' % (text, json.dumps(reply_markup, sort_keys=True))
    return text


class BotConversation:  # pylint: disable=missing-docstring,too-few-public-methods

    def __init__(self, *modules):

        def dummymod(ctx):  # pylint: disable=missing-docstring,unused-argument
            return ctx.command == 'dummymod' and ctx.reply_text('DUMMYMOD')

        self.multibot = multibot.MultiBot(set(modules) | {admin, dummymod, help})
        ntelebot.bot.Bot('1234:test').getme.respond(json={
            'ok': True,
            'result': {
                'username': 'modulestestbot'
            },
        })
        username = self.multibot.add_bot('1234:test')
        self.bot = self.multibot._build_bot(username)  # pylint: disable=protected-access
        self.bot.config['issue37']['admin']['admins'] = [1000]

    def raw_inline(self, text, user_id=1000):
        """Simulate an inline query (@BOTNAME text)."""

        user = {'id': user_id}
        inline_query = {'offset': '', 'from': user, 'query': text, 'id': user_id * 2}
        update = {'inline_query': inline_query}
        responses = []

        def _handler(request, unused_context):
            responses.append(json.loads(request.body.decode('ascii')))
            return {'ok': True, 'result': {}}

        self.bot.answer_inline_query.respond(json=_handler)
        self.multibot.dispatcher(self.bot, update)
        return responses

    # pylint: disable=too-many-arguments
    def raw_message(self,
                    text,
                    user_id=1000,
                    chat_type='private',
                    forward_date=None,
                    forward_from=None):
        """Simulate a private message."""

        user = {'id': user_id, 'username': 'user%s' % user_id, 'first_name': 'User%s' % user_id}
        if chat_type == 'private':
            chat = {'id': user_id, 'type': 'private'}
        else:
            chat = {'id': -1001000000000 - user_id, 'type': chat_type, 'title': 'Group Chat'}
        if chat_type == 'channel':
            channel_post = {
                'author_signature': user['first_name'],
                'chat': chat,
                'message_id': user_id * 2,
                'text': text,
            }
            update = {'channel_post': channel_post}
        else:
            message = {'from': user, 'chat': chat, 'message_id': user_id * 2, 'text': text}
            if forward_date:
                message['forward_date'] = forward_date
            if forward_from:
                message['forward_from'] = {'id': forward_from}
            update = {'message': message}
        responses = []

        def _handler(request, unused_context):
            response = json.loads(request.body.decode('ascii'))
            responses.append(response)
            message = {'message_id': 12345}
            if response.get('caption'):
                message['caption'] = 'CAPTION'
            return {'ok': True, 'result': message}

        self.bot.edit_message_caption.respond(json=_handler)
        self.bot.edit_message_text.respond(json=_handler)
        self.bot.forward_message.respond(json=_handler)
        self.bot.send_message.respond(json=_handler)
        self.bot.send_photo.respond(json=_handler)
        self.multibot.dispatcher(self.bot, update)
        return responses

    # pylint: enable=too-many-arguments

    @staticmethod
    def format_messages(messages):
        """Format a sequence of intercepted API calls into a transcript representation."""

        return '\n\n'.join(map(_format_message, messages))

    def message(self, *args, **kwargs):
        """Simulate a private message (with human-friendly output)."""

        return self.format_messages(self.raw_message(*args, **kwargs))


@pytest.fixture
def build_conversation():  # pylint: disable=missing-docstring
    return BotConversation



"""Tests for metabot.calendars.base (using static.Calendar instead of a local class)."""

from metabot.calendars import static


def test_client():
    """Test base.Calendar (via static.Calendar) as a consumer."""

    calendar = static.Calendar('static:dummy')
    assert calendar.last_update is None

    assert calendar.add({
        'start': 2000,
        'end': 3000,
    }) == {
        'description': '',
        'end': 3000,
        'id': '1',
        'local_id': 'c2cf0008:356a192b',
        'location': '',
        'start': 2000,
        'summary': '',
        'updated': 1001,
    }
    assert calendar.last_update == 1001

    assert calendar.update('c2cf0008:356a192b', {'location': 'Dummy\nLocation'}) == {
        'description': '',
        'end': 3000,
        'id': '1',
        'local_id': 'c2cf0008:356a192b',
        'location': 'Dummy, Location',
        'start': 2000,
        'summary': '',
        'updated': 1002,
    }
    assert calendar.last_update == 1002

    assert calendar.update('c2cf0008:356a192b', {'location': 'TBD'}) == {
        'description': '',
        'end': 3000,
        'id': '1',
        'local_id': 'c2cf0008:356a192b',
        'location': '',
        'start': 2000,
        'summary': '',
        'updated': 1003,
    }
    assert calendar.last_update == 1003

    calendar.remove('c2cf0008:356a192b')


def test_server():
    """Test base.Calendar (via static.Calendar) as a data provider."""

    calendar = static.Calendar('static:dummy')
    assert calendar.calcode == 'c2cf0008'
    assert calendar.events == {}
    assert calendar.last_update is None

    # pylint: disable=protected-access

    assert calendar._updated({
        'id': 'alpha',
        'start': 2000,
        'end': 3000,
        'updated': 4000,
    })

    assert calendar.events == {
        'c2cf0008:be76331b': {
            'description': '',
            'end': 3000,
            'id': 'alpha',
            'local_id': 'c2cf0008:be76331b',
            'location': '',
            'start': 2000,
            'summary': '',
            'updated': 4000,
        },
    }

    assert calendar._updated({
        'id': 'alpha',
        'start': 2000,
        'end': 5000,
        'updated': 6000,
    })

    assert calendar.events == {
        'c2cf0008:be76331b': {
            'description': '',
            'end': 5000,
            'id': 'alpha',
            'local_id': 'c2cf0008:be76331b',
            'location': '',
            'start': 2000,
            'summary': '',
            'updated': 6000,
        },
    }

    assert not calendar._updated({
        'id': 'alpha',
        'start': 2000,
        'end': 5000,
        'updated': 6000,
    })

    assert calendar._removed('alpha')
    assert calendar.events == {}
    assert not calendar._removed('alpha')



"""Manage the admin list."""

import uuid

from metabot.util import adminui

BOOTSTRAP_TOKEN = uuid.uuid4().hex


def modhelp(ctx, unused_modconf, sections):  # pylint: disable=missing-docstring
    bots = sorted(username for username, botconf in ctx.bot.multibot.conf['bots'].items()
                  if ctx.user['id'] in botconf['issue37']['admin']['admins'])

    if bots:
        sections['commands'].add('/admin \u2013 Manage the admin list')


def moddispatch(ctx, msg, modconf):  # pylint: disable=missing-docstring
    if ctx.type in ('message', 'callback_query'):
        if ctx.command == 'admin':
            return default(ctx, msg)
        if ctx.command == 'whoami':
            return whoami(ctx, msg)
        if ctx.command == '_bootstrap':
            return bootstrap(ctx, msg, modconf)

    return False


def default(ctx, msg):  # pylint: disable=missing-docstring
    ctx.private = True
    msg.path('/admin', 'Bot Admin')
    frame = adminui.Frame(ctx, msg, ctx.bot.multibot.conf, 'bots', None, ctx.text)
    menu = adminui.Menu()
    for username, botconf in frame.value.items():
        if ctx.user['id'] in botconf['issue37']['admin']['admins']:
            menu.add(username)

    if not menu.fields:
        return msg.add(
            "Hi! You aren't one of my admins. If you should be, ask a current admin to add you by "
            'opening a chat with me (@%s) and typing:\n'
            '\n'
            '<pre>/admin %s admin add %s</pre>', ctx.bot.username, ctx.bot.username, ctx.user['id'])

    frame, handler = menu.select(frame)
    if not handler:
        msg.add('This is a metabot! Check out https://github.com/nmlorg/metabot/issues to keep '
                'track of bugs and features.')
        try:
            msg.add(open('config/admin_motd').read().strip())
        except IOError:
            pass
        msg.add('To configure your bot, choose its username:')
        return menu.display(frame, what='bot')

    msg.path(frame.field)
    ctx.targetbotuser = frame.field
    ctx.targetbotconf = frame.value

    menu = adminui.Menu()
    for modname, module in ctx.bot.multibot.modules.items():
        if hasattr(module, 'admin'):
            menu.add(modname, module.admin, module.__doc__.splitlines()[0].rstrip('.'))

    if not menu.fields:  # pragma: no cover
        return msg.add(
            "Hi! There aren't any configurable modules installed. Contact a metabot admin to "
            'install one.')

    menu.handle(frame, what='module')


def whoami(ctx, msg):
    """Simply tell the user their userid (primarily to copy/paste into /admin BOTNAME admin)."""

    msg.add('<code>%s</code>', ctx.user['id'])


def bootstrap(ctx, msg, modconf):
    """Add the user who sent the command to the current bot's admin list."""

    if ctx.text == BOOTSTRAP_TOKEN and not modconf['admins']:
        modconf['admins'] = [ctx.user['id']]
        msg.add('Added %s to the admin list.', ctx.user['id'])


def admin(frame):  # pylint: disable=too-many-branches
    """Handle /admin BOTNAME admin (configure the admin module itself)."""

    ctx, msg, modconf = frame.ctx, frame.msg, frame.value
    if 'admins' not in modconf:  # pragma: no cover
        modconf['admins'] = []

    target = None
    if ctx.forward_from:
        target = ctx.forward_from
    elif frame.text.isdigit():
        target = int(frame.text)
    elif ctx.forwarded:
        msg.add(
            "I can't tell exactly who sent that message, possibly because the sender enabled <a "
            'href="https://telegram.org/blog/unsend-privacy-emoji#anonymous-forwarding">Anonymous '
            'Forwarding</a>. You can either ask them to temporarily re-enable normal forwarding, '
            'or ask them to open a chat with me, type <code>/whoami</code>, then forward or '
            'copy/paste the number they receive to you (to copy/paste on to me).')
    elif frame.text:
        msg.add("I'm not sure what <code>%s</code> is\u2014it's not a user id!", frame.text)

    if target in modconf['admins']:
        if target == ctx.user['id']:
            msg.add("You can't remove yourself from the admin list.")
        else:
            modconf['admins'].remove(target)
            msg.add('Removed %s from the admin list.', target)
    elif target:
        modconf['admins'].append(target)
        modconf['admins'].sort()
        msg.add('Added %s to the admin list.', target)

    msg.action = 'Choose an admin'
    msg.add('Forward a message from a user to add or remove them, or select an existing admin to '
            'remove.')

    for admin_id in sorted(modconf['admins']):
        if admin_id != ctx.user['id']:
            userinfo = ctx.bot.multibot.conf['users'].get(admin_id)
            if userinfo:
                userstr = '%s (%s)' % (userinfo['name'], admin_id)
            else:
                userstr = '%s' % admin_id
            msg.button('Remove ' + userstr, '%s' % admin_id)
